program Sprites;
{****************************************************************************
** Demonstration of scrolling a tiled background.                          **
**  by Steven H Don                                                        **
**                                                                         **
** Please note: SpriteRec is the same type as used in the sprite routines  **
** on my homepage. The MakeSprite routine and some, less significant       **
** others also originate from those files.                                 **
**                                                                         **
** For questions, feel free to e-mail me.                                  **
**                                                                         **
**    shd@earthling.net                                                    **
**    http://shd.cjb.net                                                   **
**                                                                         **
****************************************************************************}

uses Crt;

type BufferType = Array [0..63999] of Byte;

type SpriteRec = record
  {These indicate the dimensions of the sprite}
  Width, Height, Total : Word;
  {This points to the actual sprite data, it may occupy 64000 bytes, which
  is the same size as the buffer.}
  Data : ^BufferType;

  {NOTE: The following values are optional}

  {These indicate the location of the sprite on screen}
  X, Y : Word;
  {These indicate the direction of the sprite movement}
  Xdir, Ydir : ShortInt;
end;

var
  {An offscreen buffer}
  Buffer  : ^BufferType;
  {Reserve space for 16 tiles}
  Tiles   : Array [0..15] of SpriteRec;
  {The map of tiles itself}
  TileMap : Array [0..(128*128-1)] of Byte;

{Switches to VGA 320x200 256 colour}
procedure VGAScreen; assembler;
asm
  mov ax, $13
  int $10
end;

{Switches to text 80x25}
procedure TextScreen; assembler;
asm
  mov ax, $3
  int $10
end;

{Sets a pixel in the buffer to a specific colour}
procedure PutPixel (x, y : Word; c : Byte);
begin
  if (x < 320) and (y < 200) then
    Buffer^ [y shl 8 + y shl 6 + x] := c;
end;

{Reads the colour value of a pixel from the buffer}
function ReadPixel (x, y : Word) : Byte;
begin
  if (x < 320) and (y < 200) then
    ReadPixel := Buffer^ [y shl 8 + y shl 6 + x];
end;

{Assigns memory for a buffer}
function MakeBuffer : Boolean;
begin
  {Check if enough memory is available}
  if MaxAvail < 64000 then begin
    {If not, return false}
    MakeBuffer := false;
    exit;
  end;
  New (Buffer);
end;

{Clears the buffer}
procedure ClearBuffer;
begin
  fillchar (Buffer^, 64000, 0);
end;

{Displays the buffer}
procedure DisplayBuffer;
begin
  move (Buffer^,          {from buffer}
        Mem [$A000:0000], {to screen}
        64000);           {64000 pixels}
end;

function MakeSprite (var Sprite : SpriteRec; x1, y1, x2, y2 : Word) : Boolean;
var
  Size, x, y   : Word;
  CurrentPixel : Word;
  SpriteMem    : Pointer;

begin
  {Calculate amount of memory needed for sprite}
  Size := (y2 - y1 + 1) * (x2 - x1 + 1) - 1;
  {Check if enough memory is available}
  if MaxAvail < Size then begin
    {If not, return false}
    MakeSprite := false;
    exit;
  end;
  {Assign memory}
  GetMem (Sprite.Data, Size);
  {Set dimensions}
  Sprite.Width   := x2 - x1;
  Sprite.Height  := y2 - y1;
  Sprite.Total   := Size;
  {Read sprite from the screen buffer}
  CurrentPixel := 0;
  for y := y1 to y2 do begin
    for x := x1 to x2 do begin
      Sprite.Data^ [CurrentPixel] := ReadPixel (x, y);
      inc (CurrentPixel);
    end;
  end;
  {Return true}
  MakeSprite := true;
end;

procedure DrawTiledBackground (tlX, tlY : Word);
var
  {Temporary x and y variables}
  tX, tY,
  {Offset into the buffer, there's no need to use Putpixel when filling
   linearly.}
  o,
  {These give the current coordinates in the tile map.}
  TMapX, TMapY,
  {These give the coordinates of a tile in the tile map.}
  TileX, TileY,
  {These give the coordinates of a pixel in the tile.}
  TPixX, TPixY : Word;

begin
  {Start at top left of the screen}
  o := 0;
  for tY := 0 to 199 do for tX := 0 to 319 do begin
    {Calculate which pixel in the tile map is needed.}
    TMapX := tlX + tX; TMapY := tlY + tY;
    {
      Note that the following calculations assume tiles of 16x16 pixels each.
      If you change the tile size, you'll need to change the references
      to '>> 4' and '& 15' to other values.
    }
    {From that, calculate from what tile that pixel should come.}
    TileX := TMapX shr 4; TileY := TMapY shr 4;
    {And which pixel in the tile should be drawn.}
    TPixX := TMapX and 15; TPixY := TMapY and 15;
    {Draw the pixel}
    Buffer^ [o] := Tiles [TileMap [(TileY shl 7) + TileX]]
                   .Data^ [(TPixY shl 4) + TPixX];
    {Move on to next pixel}
    inc (o);
  end;
end;

var
  {Temporary x and y variables}
  tX, tY     : Word;
  {What x and y coordinates should be displayed at the top left of the screen}
  tlX, tlY   : Integer;
  {The direction the screen should bounce around the map.}
  xdir, ydir : Integer;

begin
  {Switch to graphics mode}
  VGAScreen;

  {Assign memory to buffer}
  if not MakeBuffer then begin
    {If there wasn't enough memory, quit}
    TextScreen;
    Writeln ('Not enough memory for frame buffer!');
    Halt;
  end;

  {Clear the buffer}
  ClearBuffer;

  {Set up a sprite consisting of a blue block}
  for tX := 0 to 15 do begin
    for tY := 0 to 15 do begin
      PutPixel (tX, tY, 1);
    end;
  end;
  MakeSprite (Tiles [0], 0, 0, 15, 15);

  {Set up a sprite consisting of a red rectangle}
  for tX := 0 to 15 do begin
    PutPixel (16 + tX,  0, 4);
    PutPixel (16 + tX, 15, 4);
    PutPixel (16 +  0, tX, 4);
    PutPixel (16 + 15, tX, 4);
  end;
  MakeSprite (Tiles [1], 16 + 0, 0, 16 + 15, 16 + 15);

  {Set up a tiled background of 128x128 tiles}
  for tY := 0 to 127 do for tX := 0 to 127 do
    {This generates a checkerboard...}
    TileMap [tY shl 7 + tX] := ((tY and 1) + tX) and 1;

  {Start at top left corner, moving down-right}
  tlX := 0; tlY := 0;
  xdir := 2; ydir := 2;

  while not KeyPressed do begin
    {Draw a tiled background, with coordinates tlX and tlY.}
    DrawTiledBackground (tlX, tlY);

    {Display the buffer - if you leave this out, you won't see a thing}
    DisplayBuffer;

    {Update coordinates to bounce around the screen}
    inc (tlX, xdir);
    inc (tlY, ydir);
    if (tlX < 0) then begin xdir := 1; tlX := 0; end;
    if (tlY < 0) then begin ydir := 1; tlY := 0; end;
    if (tlX > (128 * 16 - 320)) then begin xdir := -xdir; tlX := (128 * 16 - 320); end;
    if (tlY > (128 * 16 - 200)) then begin ydir := -ydir; tlY := (128 * 16 - 200); end;
  end;
  ReadKey;
end.